/*
 * FILE:    arc.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "arc.h"
#include "labelplacer.h"
#include "drawprops.h"

#include <math.h>
#include "defs.h"

#define RAD2DEG(x) ((x)/(2*M_PI)*360)
#define LINE_DISTANCE 5

Arc::Arc() {
  source = NULL;
  target = NULL;
  type = Standard;
  storeActions = 0;
  consumption = 0;
  production = 0;
}

Arc::~Arc() {
}


// Returns the list of points that make up the line to draw
// Pre:  source && target
// Post: pl.size() >= 2
QVector<QPointF> Arc::screenPointList() const {
  QVector<QPointF> pl;
  
  // Add source position
  if (points.isEmpty()) {
    pl << source->getConnectionPoint(target->getPosition());
  } else {
    pl << source->getConnectionPoint(points.first());
  }
  
  // Add all given "via's"
  foreach (QPoint p, points) {
    pl += p;
  }
  
  // Add target position
  if (points.isEmpty()) {
    // Source point only
    pl << target->getConnectionPoint(source->getPosition());
  } else {
    // Via's available
    pl << target->getConnectionPoint(points.last());
  }
  
  return pl;
}

// Pre: p.isActive() && drawProps
void Arc::draw (QPainter *p, DrawProps *drawProps) {
//  qDebug("Arc::draw() id=%s", qPrintable(id));
  
  if (source == NULL || target == NULL) {
    qDebug("No source or target for arc %s", qPrintable(id));
    return;
  }
  
  // Get list of points that make up the line to draw
  QPolygonF pointList(screenPointList());
  
  // Draw line
  if (type == Reset) {
    p->setPen(drawProps->pen("Reset"));
    p->drawPolyline(pointList);
  } else if (type == StoreAction) {
    p->setPen(drawProps->pen("Default"));
    p->drawPolyline(pointList);
  } else {
    p->setPen(drawProps->pen("Default"));
    p->setBrush(drawProps->brush("Arrow"));
    if (type == Inhibitor) {
      drawPolylineWithArrows(p, pointList,
                             drawProps->shape("Inhibitor"), QPainterPath());
    } else if (type == Standard) {
      QPainterPath a, b;
      if (consumption > 0) {
        b = drawProps->shape("Arrow");
      }
      if (production > 0) {
        a = drawProps->shape("Arrow");
      }
      drawPolylineWithArrows(p, pointList, a, b);
    }
  }
}

// Pre: pointList.size() >= 2
void Arc::drawPolylineWithArrows(QPainter *p, QPolygonF pointList,
    QPainterPath startArrow, QPainterPath endArrow) {
  // Draw lines
  QPointF firstPoint = pointList.first();
  QPointF endPoint = pointList.last();
  if (!startArrow.isEmpty()) {
    // Move first point in pointList towards second point
    QLineF line(firstPoint, pointList[1]);
    qreal arrowWidth = startArrow.controlPointRect().width();
    pointList[0] = line.pointAt(arrowWidth / line.length());
  }
  if (!endArrow.isEmpty()) {
    // Move last point in pointList towards second last point
    QLineF line(endPoint, pointList[pointList.size() - 2]);
    qreal arrowWidth = endArrow.controlPointRect().width();
    pointList[pointList.size() - 1] = line.pointAt(arrowWidth / line.length());
  }
  p->drawPolyline(pointList);
  
  // Draw arrows
  p->setPen(Qt::NoPen);
  if (!startArrow.isEmpty()) {
    drawArrow(p, startArrow, pointList[1], firstPoint);
  }
  if (!endArrow.isEmpty()) {
    drawArrow(p, endArrow, pointList[pointList.size() - 2], endPoint);
  }
}

void Arc::drawArrow(QPainter *p, QPainterPath arrow, QPointF from, QPointF to) {
  QMatrix m, orig = p->matrix();
  m.translate(to.x(), to.y());

  QPointF q(from - to);
  if (q.x() == 0) {
    if (q.y() > 0) {
      m.rotate(90);
    } else {
      m.rotate(270);
    }
  } else if (q.x() > 0) {
    m.rotate(RAD2DEG(atan(1.0 * q.y() / q.x())));
  } else { // q.x() < 0
    m.rotate(180.0 + RAD2DEG(atan(1.0 * q.y() / q.x())));
  }

  // Draw arrowhead
  p->setMatrix(m, true);
  p->drawPath(arrow);
  p->setMatrix(orig, false);
}

// Pre: labelPlacer
void Arc::updateLabels(LabelPlacer *labelPlacer, DrawProps *) {
  if (source == NULL || target == NULL) {
    qDebug("No source or target for arc %s", qPrintable(id));
    return;
  }
  
  // Get list of points that make up the line to draw
  QVector<QPointF> pointList(screenPointList());
  
  // Let the label placer know about these lines
  for (int i = 1; i < pointList.size(); i++) {
    labelPlacer->addLine(QLine(pointList.at(i-1).toPoint(),
                               pointList.at(i).toPoint()));
    
  }
  
  // Make our label
  QString label;
  if (storeActions & ActionC) {
    label += "C";
  }
  if (storeActions & ActionR) {
    label += "R";
  }
  if (storeActions & ActionU) {
    label += "U";
  }
  if (storeActions & ActionD) {
    label += "D";
  }
  if (label.length() > 0) {
    label = QString("(%1)").arg(label);
    if (name.length() > 0) {
      label = name + "\n" + label;
    }
  } else if (name.length() > 0) {
    label = name;
  }

  if (label.length() > 0 || consumption > 1 || production > 1) {
    // Find middle of arc and place label there
    QPointF labelPos;
    if (pointList.size() % 2 == 0) {
      // Odd number of lines, draw text near line no. pointList.size()/2 - 1
      // { pointList.size() >= 2 guaranteed by screenPointList() }
      labelPos = (pointList.at(pointList.size()/2 - 1) +
                  pointList.at(pointList.size()/2)) / 2;
    } else {
      // Odd number of points, draw text near point no. (pointList.size()-1)/2
      // { pointList.size() >= 2 guaranteed by screenPointList() }
      labelPos = pointList.at((pointList.size()-1)/2);
    }
    
    if (label.length() > 0) {
      labelPlacer->addLabel(labelPos.toPoint(), label);
    }
    
    // Tokenflow label
    bool pc = consumption > 1 || (production > 1 && consumption > 0);
    bool pp = production > 1 || (consumption > 1 && production > 0);
    
    if (pc || pp) {
      label = "(";
      if (pc) {
        label += /*(pp?"C:":"") +*/ QString::number(consumption);
      }
      if (pc && pp) {
        label += ", ";
      }
      if (pp) {
        label += /*(pc?"P:":"") +*/ QString::number(production);
      }
      label += ")";
      
      labelPlacer->addLabel(labelPos.toPoint(), label);
    }
  }
}

bool Arc::containsPos(QPoint pos) const {
  QVector<QPointF> pl = screenPointList();
  for (int i = 1; i < pl.size(); i++) {
    QPointF p1 = pl.at(i-1);
    QPointF p2 = pl.at(i);
    
    // TODO: find a nicer solution if possible
    p2 = p2 - p1;
    QPointF q(QPointF(pos) - p1);
    QMatrix m;
    if (static_cast<int>(p2.x()) == 0) {
      if (p2.y() > 0) {
        m.rotate(-90);
      } else {
        m.rotate(-270);
      }
    } else if (p2.x() > 0) {
      m.rotate(-RAD2DEG(atan(1.0 * p2.y() / p2.x())));
    } else { // (p2.x() < 0)
      m.rotate(-180-RAD2DEG(atan(1.0 * p2.y() / p2.x())));
    }
    q = q * m;
    p2 = p2 * m;
    if (q.x() >= 0 &&
        q.x() <= p2.x() &&
        fabs(q.y()) < LINE_DISTANCE) {
      return true;
    }
  }
  return false;
}

QRect Arc::getBoundingRect() {
  QRect bb;
  QPointF curP;
  foreach(curP, points) {
    if (bb.isNull()) {
      bb.setTopLeft(curP.toPoint());
      bb.setWidth(1);
      bb.setHeight(1);
    } else {
      bb |= QRect(curP.toPoint(), QSize(1,1));
    }
  }
  return bb;
}

const QString Arc::getToolTip() const {
  QString text = name;
  if (text.isEmpty()) {
    text = id;
  } else {
    text.replace('<', "&lt;");
    text.replace('>', "&gt;");
    text = QString("<b>") + text + "</b> (" + id +")";
  }
  return text;
}

// Pre: other && other->getType() == this->getType()
void Arc::merge(Arc *other, bool swapTargets) {
  if (!other->getName().isEmpty()) {
    if (name.isEmpty()) {
      name = other->getName();
    } else {
      name += ", " + other->getName();
    }
  }
  id += "," + other->getId();
  
  switch (type) {
    case Standard:
      if (!swapTargets) {
        updateProduction(other->getProduction());
        updateConsumption(other->getConsumption());
      } else {
        updateConsumption(other->getProduction());
        updateProduction(other->getConsumption());
      }
      break;
    case StoreAction:
      storeActions |= other->getStoreAction();
      break;
    case Reset:
    case Inhibitor:
      // Nothing to be done, double reset or double inhibitor arcs
      // will be ignored
      break;
  }
}

const QVector<QPoint> Arc::getPoints() {
  return points;
}

void Arc::setPoints(QVector<QPoint> value) {
  points = value;
}

void Arc::addPoint (QPoint value) {
  points << value;
}

