/*
 * FILE:    drawprops.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "drawprops.h"

#include <QPen>
#include <QBrush>
#include <QPen>
#include <QString>
#include <QSettings>


DrawProps::DrawProps() {
  m_zoom = 1.0;
}

DrawProps::~DrawProps() {
}

QPen DrawProps::pen(QString name) const {
  if (pens.contains(name)) {
    return pens[name];
  } else {
    qWarning("DrawProps::pen() Unknown pen: ``%s''", qPrintable(name));
    return QPen(Qt::black);
  }
}

void DrawProps::setPen(QString name, QPen value) {
  pens[name] = value;
}

QBrush DrawProps::brush(QString name) const {
  if (brushes.contains(name)) {
    return brushes[name];
  } else {
    qWarning("DrawProps::brush() Unknown brush: ``%s''", qPrintable(name));
    return Qt::NoBrush;
  }
}

void DrawProps::setBrush(QString name, QBrush value) {
  brushes[name] = value;
}

QFont DrawProps::font(QString name) const {
  if (fonts.contains(name)) {
    return fonts[name];
  } else {
    qWarning("DrawProps::font() Unknown font: ``%s''", qPrintable(name));
    return QFont();
  }
}

void DrawProps::setFont(QString name, QFont value) {
  fonts[name] = value;
}

bool DrawProps::option(QString name) const {
  if (options.contains(name)) {
    return options[name];
  } else {
    qWarning("DrawProps::option() Unknown option: ``%s''", qPrintable(name));
    return false;
  }
}

void DrawProps::setOption(QString name, bool value) {
  options[name] = value;
}

QString DrawProps::string(QString name) const {
  if (strings.contains(name)) {
    return strings[name];
  } else {
    qWarning("DrawProps::string() Unknown string: ``%s''", qPrintable(name));
    return QString();
  }
}

void DrawProps::setString(QString name, QString value) {
  strings[name] = value;
}

QPainterPath DrawProps::shape(QString name) const {
  if (shapes.contains(name)) {
    return shapes[name];
  } else {
    qWarning("DrawProps::shape() Unknown shape: ``%s''", qPrintable(name));
    return QPainterPath();
  }
}

double DrawProps::zoom() const {
  return m_zoom;
}

void DrawProps::setZoom(double value) {
  m_zoom = value;
}

QStringList DrawProps::penNames() const {
  return pens.keys();
}

QStringList DrawProps::brushNames() const {
  return brushes.keys();
}

QStringList DrawProps::fontNames() const {
  return fonts.keys();
}

QStringList DrawProps::optionNames() const {
  return options.keys();
}

void DrawProps::resetAll() {
  resetPens();
  resetBrushes();
  resetFonts();
  resetOptions();
  resetStrings();
  resetShapes();
}

void DrawProps::resetPens() {
  QPen pen(Qt::black);
  pen.setWidth(1);
  pen.setJoinStyle(Qt::MiterJoin);
  pen.setCapStyle(Qt::FlatCap);
  
  setPen("Default", pen);
  setPen("Text", pen);
  
  // Using customized pattern to make it look more like Yasper
  QVector<qreal> dashPattern;
  dashPattern << 5.0 << 5.0;
  pen.setDashPattern(dashPattern);
  setPen("Reset", pen);
  
  pen.setStyle(Qt::SolidLine);
  pen.setColor(QColor(47, 79, 79)); // Dark green
  pen.setWidth(3);
  setPen("InlinePage", pen);
}

void DrawProps::resetBrushes() {
  setBrush("Default", Qt::NoBrush);
  setBrush("InlinePage", Qt::NoBrush);
  setBrush("Background", QBrush(Qt::white));
  setBrush("Arrow", QBrush(Qt::black));
  setBrush("Token", QBrush(Qt::black));
  setBrush("Shadow", QBrush(QColor(0, 0, 0, 60)));
  setBrush("Xor", QBrush(QColor(246,125,0))); // Orange
  setBrush("Page", QBrush(QColor(103,170,203))); // Blue
  setBrush("Transition", QBrush(QColor(153,204,0))); // Green
  setBrush("Emitter", QBrush(Qt::yellow));
  setBrush("Collector", QBrush(Qt::yellow));
  setBrush("ColoredPlace", QBrush(Qt::yellow));
  setBrush("ColoredStore", QBrush(Qt::yellow));
  setBrush("UncoloredPlace", QBrush(Qt::white));
  setBrush("UncoloredStore", QBrush(QColor(237, 232, 203))); // Beige
}

void DrawProps::resetFonts() {
  QFont font;
  font.setPixelSize(11);
  font.setWeight(QFont::Normal);
  setFont("Name", font);
  font.setStyle(QFont::StyleItalic);
  setFont("Role", font);
}

void DrawProps::resetOptions() {
  setOption("AntiAlias", true);
  setOption("ShowNames", true);
  setOption("ShowRoles", true);
  setOption("ShowIdsTransitions", false);
  setOption("ShowIdsPlaces", false);
  setOption("ExportTransparency", true);
}

void DrawProps::resetStrings() {
  setString("HiddenRoles", "all");
}

void DrawProps::resetShapes() {
  // Inhibitor dot
  QPainterPath inhibitor;
  inhibitor.addEllipse(0, -INHIBITOR_SIZE,
                       INHIBITOR_SIZE*2, INHIBITOR_SIZE*2);
  shapes["Inhibitor"] = inhibitor;
  
  // Arrowhead
  QPainterPath arrow;
  arrow.moveTo(0, 0);
  arrow.lineTo(9, 3);
  arrow.lineTo(9, -3);
  arrow.closeSubpath();
  shapes["Arrow"] = arrow;
}

void DrawProps::loadSettings() {
  QSettings settings;
  
  // First load the defaults
  resetAll();
  
  // Now load the settings
  /*
  // Pens are not customizable yet
  // Pens
  for (QMap<QString, QPen>::const_iterator it = pens.constBegin(); it != pens.constEnd(); ++it) {
    setPen(it.key(), settings.value(QString("Pen/") + it.key(), it.value()).value<QPen>());
  }
  */
  
  // Brushes
  for (QMap<QString, QBrush>::const_iterator it = brushes.constBegin(); it != brushes.constEnd(); ++it) {
    setBrush(it.key(), settings.value(QString("Brush/") + it.key(), it.value()).value<QBrush>());
  }
  
  // Fonts
  for (QMap<QString, QFont>::const_iterator it = fonts.constBegin(); it != fonts.constEnd(); ++it) {
    setFont(it.key(), settings.value(QString("Font/") + it.key(), it.value()).value<QFont>());
  }
  
  // Options
  for (QMap<QString, bool>::const_iterator it = options.constBegin(); it != options.constEnd(); ++it) {
    setOption(it.key(), settings.value(QString("Option/") + it.key(), it.value()).toBool());
  }
  
  // Strings
  for (QMap<QString, QString>::const_iterator it = strings.constBegin(); it != strings.constEnd(); ++it) {
    setString(it.key(), settings.value(QString("String/") + it.key(), it.value()).toString());
  }
}

void DrawProps::saveSettings() const {
  QSettings settings;
  /*
  // Save pens
  for (QMap<QString, QPen>::const_iterator it = pens.constBegin(); it != pens.constEnd(); ++it) {
    settings.setValue(QString("Pen/") + it.key(), it.value());
  }
  */
  
  // Save brushes
  for (QMap<QString, QBrush>::const_iterator it = brushes.constBegin(); it != brushes.constEnd(); ++it) {
    settings.setValue(QString("Brush/") + it.key(), it.value());
  }
  
  // Save fonts
  for (QMap<QString, QFont>::const_iterator it = fonts.constBegin(); it != fonts.constEnd(); ++it) {
    settings.setValue(QString("Font/") + it.key(), it.value());
  }
  
  // Save options
  for (QMap<QString, bool>::const_iterator it = options.constBegin(); it != options.constEnd(); ++it) {
    settings.setValue(QString("Option/") + it.key(), it.value());
  }
  
  // Save strings
  for (QMap<QString, QString>::const_iterator it = strings.constBegin(); it != strings.constEnd(); ++it) {
    settings.setValue(QString("String/") + it.key(), it.value());
  }
}

// Set draw properties such that printing (and pdf export) won't fail
void DrawProps::setPrintSafe() {
  if (QString(qVersion()) == "4.1.0") {
    pens["Reset"].setStyle(Qt::DashLine);
  }
  brushes["Background"] = Qt::NoBrush;
  brushes["Shadow"] = QBrush(QColor(195, 195, 195)); // Light gray
}
