/*
 * FILE:    labelplacer.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "labelplacer.h"
#include "drawprops.h"

LabelPlacer::LabelPlacer() {
  hasChanged = false;
}

LabelPlacer::~LabelPlacer() {
}

// Pre: drawProps && (painter ? painter.isActive() : true)
void LabelPlacer::place(DrawProps *drawProps, QPainter *painter /*= NULL*/) {
  //qDebug("LabelPlacer::place()");
  
  // Only place if there was an update
  if (!hasChanged) {
    return;
  }
  
  placedLabels.clear();
  placedRoleLabels.clear();
  
  QImage *im = NULL;
  bool cleanupPainter = false;
  if (!painter) {
    // Make temporary image and painter to be able to get the bounding rects...
    // TODO: use QFontMetrics
    im = new QImage(1, 1, QImage::Format_RGB32);
    painter = new QPainter(im);
    cleanupPainter = true;
  }
  
  painter->setFont(drawProps->font("Name"));
  QTextOption opt;
  opt.setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
  opt.setWrapMode(QTextOption::NoWrap);
  
  QPair<QPoint, QString> pair;
  foreach (pair, labels) {
    // Find bouding box of text
    QRect bb(pair.first.x(), pair.first.y(), 1, 1);
    bb = painter->boundingRect(bb, pair.second, opt).toRect();

    QRect problem = findRect(bb);

    if (problem.isNull()) {
      if (!findLine(bb).isNull()) {
        //qDebug("Only found a line problem for label ``%s'' at pos. (%d, %d)", qPrintable(pair.second), pair.first.x(), pair.first.y());
        problem.setTopLeft(pair.first);
        problem.setWidth(1);
        problem.setHeight(1);
      }/* else {
        qDebug("Found no problem for label ``%s'' at pos. (%d, %d)", qPrintable(pair.second), pair.first.x(), pair.first.y());
      }*/
    }


    // Main placement algorithm (for normal labels)
    QRect fr(bb);
    if (!problem.isNull()) {
      QVector<QRect> positions;
      // Bottom
      fr = bb; fr.moveTop(problem.bottom() + LABEL_DISTANCE);
      positions.append(fr);
      // Top
      fr = bb; fr.moveBottom(problem.top() - LABEL_DISTANCE);
      positions.append(fr);
      // Left
      fr = bb; fr.moveRight(problem.left() - LABEL_DISTANCE);
      positions.append(fr);
      // Right
      fr = bb; fr.moveLeft(problem.right() + LABEL_DISTANCE);
      positions.append(fr);
      // Bottom-left
      fr = bb;
      fr.moveTop(problem.bottom() + LABEL_DISTANCE);
      fr.moveRight(problem.left() - LABEL_DISTANCE);
      positions.append(fr);
      // Bottom-right
      fr = bb;
      fr.moveTop(problem.bottom() + LABEL_DISTANCE);
      fr.moveLeft(problem.right() + LABEL_DISTANCE);
      positions.append(fr);
      // Top-right
      fr = bb;
      fr.moveBottom(problem.top() - LABEL_DISTANCE);
      fr.moveLeft(problem.right() + LABEL_DISTANCE);
      positions.append(fr);
      // Top-left
      fr = bb;
      fr.moveBottom(problem.top() - LABEL_DISTANCE);
      fr.moveRight(problem.left() - LABEL_DISTANCE);
      positions.append(fr);
      
      
      bool ok = false;
      foreach (fr, positions) {
        if (findLine(fr).isNull() && findRect(fr).isNull()) {
          ok = true;
          break;
        }
      }
      if (!ok) {
        qDebug("Label ``%s'' could not be placed", qPrintable(pair.second));
        fr = positions.at(0);
      }
    }

    placedLabels.append(qMakePair(fr, pair.second));
  }
  
  painter->setFont(drawProps->font("Role"));
  // TODO: this ugly copy
  foreach (pair, roleLabels) {
    // Find bouding box of text
    QRect bb(pair.first.x(), pair.first.y(), 1, 1);
    bb = painter->boundingRect(bb, pair.second, opt).toRect();

    QRect problem = findRect(bb);

    if (problem.isNull()) {
      if (!findLine(bb).isNull()) {
        qDebug("Only found a line problem for label ``%s'' at pos. (%d, %d)", qPrintable(pair.second), pair.first.x(), pair.first.y());
        problem.setTopLeft(pair.first);
        problem.setWidth(1);
        problem.setHeight(1);
      } else {
        qDebug("Found no problem for label ``%s'' at pos. (%d, %d)", qPrintable(pair.second), pair.first.x(), pair.first.y());
      }
    }


    // Main placement algorithm (for role labels)
    QRect fr(bb);
    if (!problem.isNull()) {
      QVector<QRect> positions;
      // Bottom
      fr = bb; fr.moveTop(problem.bottom() + LABEL_DISTANCE);
      positions.append(fr);
      // Top
      fr = bb; fr.moveBottom(problem.top() - LABEL_DISTANCE);
      positions.append(fr);
      // Left
      fr = bb; fr.moveRight(problem.left() - LABEL_DISTANCE);
      positions.append(fr);
      // Right
      fr = bb; fr.moveLeft(problem.right() + LABEL_DISTANCE);
      positions.append(fr);
      // Bottom-left
      fr = bb;
      fr.moveTop(problem.bottom() + LABEL_DISTANCE);
      fr.moveRight(problem.left() - LABEL_DISTANCE);
      positions.append(fr);
      // Bottom-right
      fr = bb;
      fr.moveTop(problem.bottom() + LABEL_DISTANCE);
      fr.moveLeft(problem.right() + LABEL_DISTANCE);
      positions.append(fr);
      // Top-right
      fr = bb;
      fr.moveBottom(problem.top() - LABEL_DISTANCE);
      fr.moveLeft(problem.right() + LABEL_DISTANCE);
      positions.append(fr);
      // Top-left
      fr = bb;
      fr.moveBottom(problem.top() - LABEL_DISTANCE);
      fr.moveRight(problem.left() - LABEL_DISTANCE);
      positions.append(fr);
      
      
      bool ok = false;
      foreach (fr, positions) {
        if (!isObstructed(fr)) {
          ok = true;
          break;
        }
      }
      if (!ok) {
        qDebug("Label ``%s'' could not be placed", qPrintable(pair.second));
        fr = positions.at(0);
      }
    }

    placedRoleLabels.append(qMakePair(fr, pair.second));
  }
  
  if (cleanupPainter) {
    delete painter;
    delete im;
    painter = NULL;
    im = NULL;
  }
  
  hasChanged = false;
}

// Pre: p.isActive() && drawProps
void LabelPlacer::draw(QPainter *p, DrawProps *drawProps) {
  //qDebug("LabelPlacer::draw()");
  
  place(drawProps, p); // ensure labels are placed
  
  p->setPen(drawProps->pen("Text"));
  p->setFont(drawProps->font("Name"));

  QTextOption opt;
  opt.setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
  opt.setWrapMode(QTextOption::NoWrap);
  
  QPair<QRect, QString> pair;
  // Draw normal labels
  if (drawProps->option("ShowNames")) {
    foreach (pair, placedLabels) {
      p->drawText(pair.first, pair.second, opt);
    }
  }
  
  // Draw role labels
  if (drawProps->option("ShowRoles")) {
    p->setFont(drawProps->font("Role"));
    foreach (pair, placedRoleLabels) {
      p->drawText(pair.first, pair.second, opt);
    }
  }
}

// Pre: drawProps && (painter ? painter.isActive() : true)
QRect LabelPlacer::getBoundingRect(DrawProps *drawProps, QPainter *painter /*= NULL*/) {
  qDebug("LabelPlacer::genBoundingRect()");
  
  place(drawProps, painter); // ensure labels are placed

  QRect bb;
  QPair<QRect, QString> label;
  foreach (label, placedLabels) {
    bb |= label.first;
  }
  foreach (label, placedRoleLabels) {
    bb |= label.first;
  }
  return bb;
}

void LabelPlacer::addLabel(QPoint pos, QString text) {
  if (!text.isEmpty()) {
    labels << QPair<QPoint, QString>(pos, text);
    hasChanged = true;
  }
}

void LabelPlacer::addRoleLabel(QPoint pos, QString text) {
  if (!text.isEmpty()) {
    roleLabels << QPair<QPoint, QString>(pos, text);
    hasChanged = true;
  }
}

void LabelPlacer::addLine(QLine line) {
  if (!line.isNull()) {
    lines << line;
    hasChanged = true;
  }
}

void LabelPlacer::addRect(QRect rect) {
  if (!rect.isNull()) {
    rects << rect;
    hasChanged = true;
  }
}

void LabelPlacer::clearAll() {
  labels.clear();
  placedLabels.clear();
  roleLabels.clear();
  placedRoleLabels.clear();
  lines.clear();
  rects.clear();
  hasChanged = true;
}

QRect LabelPlacer::findRect(QRect rect) {
  QRect rectI, bb;
  foreach (rectI, rects) {
    if (rectI.intersects(rect)) {
      bb |= rectI;
    }
  }
  QPair<QRect, QString> label;
  foreach (label, placedLabels) {
    if (label.first.intersects(rect)) {
      bb |= label.first;
    }
  }
  foreach (label, placedRoleLabels) {
    if (label.first.intersects(rect)) {
      bb |= label.first;
    }
  }
  return bb;
}

bool LabelPlacer::isObstructed(QRect rect) {
  QRect rectI;
  foreach (rectI, rects) {
    if (rectI.intersects(rect)) {
      return true;
    }
  }
  QPair<QRect, QString> label;
  foreach (label, placedLabels) {
    if (label.first.intersects(rect)) {
      return true;
    }
  }
  foreach (label, placedRoleLabels) {
    if (label.first.intersects(rect)) {
      return true;
    }
  }
  return !findLine(rect).isNull();
}

#define TOP     8
#define BOTTOM  4
#define RIGHT   2
#define LEFT    1
#define OUTCODE(p) ((((p).y() < rect.top())?TOP:(((p).y() >= rect.bottom())?BOTTOM:0))|(((p).x() >= rect.right())?RIGHT:(((p).x() < rect.left())?LEFT:0)))
QLine LabelPlacer::findLine(QRect rect) {
  QLine line;
  foreach (line, lines) {
    // Cohen-Sutherland algo (partly)
    int oc0 = OUTCODE(line.p1());
    int oc1 = OUTCODE(line.p2());
    if ((oc0 & oc1) == 0) {
      if (oc0 == 0 || oc1 == 0) {
        /*
        qDebug("Line ((%d, %d), (%d, %d)) intersects(a) rect ((%d, %d), (%d, %d)).",
            line.p1().x(), line.p1().y(), line.p2().x(), line.p2().y(),
            rect.left(), rect.top(), rect.right(), rect.bottom());
        */
        return line;
      }
      /*
      if ((oc0 == TOP || oc0 == BOTTOM || oc0 == RIGHT || oc0 == LEFT) &&
          (oc1 == TOP || oc1 == BOTTOM || oc1 == RIGHT || oc1 == LEFT)) {
        
        qDebug("Line ((%d, %d), (%d, %d)) intersects(b) rect ((%d, %d), (%d, %d)).",
            line.p1().x(), line.p1().y(), line.p2().x(), line.p2().y(),
            rect.left(), rect.top(), rect.right(), rect.bottom());
        
        return line;
      }
      */
      // This could perhaps be done nicer (faster?), but it works
      // if you check for intersection of "line" with 3 of the 4 sides
      // of "rect"
      QLineF rLine(rect.topLeft(), rect.topRight());
      if (rLine.intersect(line, NULL) == QLineF::BoundedIntersection) {
        return line;
      }
      rLine = QLineF(rect.topRight(), rect.bottomRight());
      if (rLine.intersect(line, NULL) == QLineF::BoundedIntersection) {
        return line;
      }
      rLine = QLineF(rect.bottomRight(), rect.bottomLeft());
      if (rLine.intersect(line, NULL) == QLineF::BoundedIntersection) {
        return line;
      }
    }
  }
  return QLine();
}
