/*
 * FILE:    mainwindow.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "mainwindow.h"
#include "batchconvertdialog.h"
#include ".ui/ui_versiondialog.h"
#include "version.h"
#include "prefdialog.h"
#include "drawwidget.h"
#include "pnmlhandler.h"
#include "net.h"
#include "drawprops.h"
#include "export.h"

#include <QPrinter>
#include <QSettings>
#include <QCloseEvent>
#include <QScrollArea>
#include <QFileDialog>
#include <QMessageBox>
#include <QPrintDialog>
#include <QImageWriter>

MainWindow::MainWindow(QWidget * parent /* = 0*/, Qt::WFlags f /* = 0*/) :
  QMainWindow(parent, f) {
  net = NULL;
  currentFile = "";
  
  // Setup UI
  ui.setupUi(this);
  
  // Main draw area
  drawArea = new DrawWidget(&drawProps);
  scrollArea = new QScrollArea(this);
  scrollArea->setBackgroundRole(QPalette::Base);
  scrollArea->setWidget(drawArea);
  setCentralWidget(scrollArea);
  
  // Setup some actions
  // Add toggle view actions to the View menu
  ui.menuView->addAction(ui.dockWidget->toggleViewAction());
  ui.menuView->addSeparator();
  ui.menuView->addAction(ui.toolBar->toggleViewAction());
  // Add compile-time-optional export actions
#ifdef QT3_SUPPORT
  ui.menuVector->addAction(ui.actionExportSVG);
#endif
  // Disable most of the actions, because no file is loaded
  setActionsEnabled(false);
  
  // Load settings
  QSettings settings;
  restoreState(settings.value("General/MainWindowState").toByteArray());
  resize(settings.value("General/MainWindowSize", QSize(800, 600)).toSize());
  lastPath = settings.value("General/LastPath", QDir::currentPath()).toString();
  lastExportPath = settings.value("General/LastExportPath", QDir::currentPath()).toString();
  lastImageFormat = settings.value("General/LastImageFormat", "png").toString();
  ui.toolTipsCheckBox->setChecked(settings.value("General/ShowToolTips", false).toBool());
  
  // Load drawProps settings
  drawProps.loadSettings();
  
  ui.namesCheckBox->setChecked(drawProps.option("ShowNames"));
  ui.rolesCheckBox->setChecked(drawProps.option("ShowRoles"));
  
  // Set scrollArea background
  QPalette sp(scrollArea->palette());
  sp.setBrush(QPalette::Base, drawProps.brush("Background"));
  scrollArea->setPalette(sp);
  
  // Connect some signals & slots
  connect(ui.actionQuit, SIGNAL(triggered()), this, SLOT(close()));
}

MainWindow::~MainWindow() {
  delete net;
}

void MainWindow::openFile(QString fileName) {
  // Create a handler for PNML files
  PnmlHandler handler;
  QXmlSimpleReader reader;
  reader.setContentHandler(&handler);
  reader.setErrorHandler(&handler);
  
  // Test if the file can be opened
  QFile file(fileName);
  if (!file.open(QFile::ReadOnly | QFile::Text)) {
    QMessageBox::warning(this, "PNML view",
                         QString("Cannot read file %1:\n%2.")
                         .arg(fileName)
                         .arg(file.errorString()));
    return;
  }
  
  // Start the parsing procedure
  QXmlInputSource xmlInputSource(&file);
  if (!reader.parse(xmlInputSource)) {
    // Parse error occured
    QMessageBox::warning(this,
                         "PNML parse error",
                         handler.fatalErrorString());
    return;
  }
  
  // File parsed and ok

  // Show message for 2 sec.
  statusBar()->showMessage("File loaded", 2000);

  // Update window title
  QFileInfo fi(fileName);
  setWindowTitle(fi.fileName() + " - PNML view");
  setActionsEnabled(true);
  currentFile = fileName;
  currentFileLastModified = fi.lastModified();

  // Remember path
  lastPath = fi.absolutePath();

  // Update net
  drawArea->setNet(handler.getNet());
  delete net;
  net = handler.getNet();
  
  if (net) {
    // Update hierarchy
    ui.hierarchyTreeWidget->clear();
    QTreeWidgetItem *root = new QTreeWidgetItem(ui.hierarchyTreeWidget);
    QString netName = handler.getNetName();
    if (netName.isNull()) {
      netName = "(root)";
    }
    root->setText(0, netName);
    QVariant varNet;
    varNet.setValue<void *>(static_cast<void *>(net));
    root->setData(0, Qt::UserRole, varNet);
    ui.hierarchyTreeWidget->setItemExpanded(root, true);
    net->updateHierarchy(root);
  }
}
  
// Enable/disable actions that are only available if a file is open
void MainWindow::setActionsEnabled(bool value) {
  ui.actionExportPDF->setEnabled(value);
  ui.actionExportEPS->setEnabled(value);
  ui.actionExportSVG->setEnabled(value);
  ui.actionExportRaster->setEnabled(value);
  ui.actionPrint->setEnabled(value);
  ui.actionZoomIn->setEnabled(value);
  ui.actionZoomOut->setEnabled(value);
  ui.actionZoomFull->setEnabled(value);
}

void MainWindow::closeEvent(QCloseEvent *e) {
  // Save settings
  QSettings settings;
  settings.setValue("General/MainWindowState", saveState());
  settings.setValue("General/MainWindowSize", size());
  settings.setValue("General/LastPath", lastPath);
  settings.setValue("General/LastExportPath", lastExportPath);
  settings.setValue("General/LastImageFormat", lastImageFormat);
  settings.setValue("General/ShowToolTips", ui.toolTipsCheckBox->isChecked());
  drawProps.saveSettings();
  e->accept();
}

void MainWindow::enterEvent(QEvent *) {
  if (!currentFile.isEmpty()) {
    QFileInfo fi(currentFile);
    if (currentFileLastModified != fi.lastModified()) {
      if (QMessageBox::question(this,
                                "File modified",
                                QString("%1 has been modified. Reload?").arg(fi.fileName()),
                                "Reload",
                                "Cancel") == 0) {
        openFile(currentFile);
      }
      // Be sure to update the last modified data, so that the user
      // won't be notified again if he clicks cancel
      currentFileLastModified = fi.lastModified();
    }
  }
}

// Update background color & redraw complete picture
void MainWindow::updateAll() {
  QPalette sp(scrollArea->palette());
  sp.setBrush(QPalette::Base, drawProps.brush("Background"));
  scrollArea->setPalette(sp);

  if (net) {
    net->setDirty(Net::AllDirty);
  }
  drawArea->sizeChanged();
}

void MainWindow::on_actionOpen_triggered() {
  // Ask user for file to open
  QString fileName =
      QFileDialog::getOpenFileName(this,
                                   "Choose a PNML file",
                                   lastPath,
                                   "PNML files (*.pnml)");
  if (fileName.isEmpty()) {
    return;
  }
  openFile(fileName);
}

void MainWindow::on_actionPreferences_triggered() {
  PrefDialog dlg(&drawProps, this);
  connect(&dlg, SIGNAL(applied()), this, SLOT(updateAll()));
  if (dlg.exec()) {
    updateAll();
  }
}

void MainWindow::on_actionPrint_triggered() {
#ifndef QT_NO_PRINTER
  if (drawArea->getNet()) {
    QPrinter printer(QPrinter::HighResolution);
    QPrintDialog *dlg = new QPrintDialog(&printer, this);
    if (dlg->exec() == QDialog::Accepted) {
      DrawProps dp(drawProps);
      dp.setPrintSafe();
      // Scale to paper
      QSize netSize = drawArea->getNet()->getSize(&dp);
      if (1.0 * netSize.width() / netSize.height() >
          1.0 * printer.width() / printer.height()) {
        // Net is wider than paper
        dp.setZoom(1.0 * printer.width() / netSize.width());
      } else {
        // Net is higher than paper
        dp.setZoom(1.0 * printer.height() / netSize.height());
      }
      
      drawArea->getNet()->draw(&printer, &dp, true);
    
      // Show message for 2 sec.
      statusBar()->showMessage("Image printed", 2000);
    }
    delete dlg;
  }
#endif
}

void MainWindow::on_actionExportPDF_triggered() {
#ifndef QT_NO_PRINTER
  if (drawArea->getNet()) {
    QString fileName(lastExportPath);
    if (!fileName.endsWith('/')) {
      fileName += "/";
    }
    QFileInfo fic(currentFile);
    fileName += fic.baseName() + ".pdf";
    fileName = QFileDialog::getSaveFileName(this,
                                            "Export PDF",
                                            fileName,
                                            "PDF file (*.pdf)");
    if (fileName.isEmpty()) {
      return;
    }
    
    bool exportOk = Export::savePdf(drawArea->getNet(), &drawProps, fileName);
    
    // Remember last export path
    QFileInfo fi(fileName);
    lastExportPath = fi.absolutePath();
    
    // Show message for 2 sec.
    if (exportOk) {
      statusBar()->showMessage("Image exported", 2000);
    } else {
      QMessageBox::warning(this, "PNML view",
                           QString("Exporting to file ``%1'' failed.")
                           .arg(fileName));
    }
  }
#endif
}

void MainWindow::on_actionExportEPS_triggered() {
#ifndef QT_NO_PRINTER
  if (drawArea->getNet()) {
    QString fileName(lastExportPath);
    if (!fileName.endsWith('/')) {
      fileName += "/";
    }
    QFileInfo fic(currentFile);
    fileName += fic.baseName() + ".eps";
    fileName = QFileDialog::getSaveFileName(this,
                                            "Export EPS",
                                            fileName,
                                            "Encapsulated PostScript (*.eps)");
    if (fileName.isEmpty()) {
      return;
    }
    
    bool exportOk = Export::saveEps(drawArea->getNet(), &drawProps, fileName);
    
    // Remember last export path
    QFileInfo fi(fileName);
    lastExportPath = fi.absolutePath();
    
    // Show message for 2 sec.
    if (exportOk) {
      statusBar()->showMessage("Image exported", 2000);
    } else {
      QMessageBox::warning(this, "PNML view",
                           QString("Exporting to file ``%1'' failed.")
                           .arg(fileName));
    }
  }
#endif
}

#ifdef QT3_SUPPORT
void MainWindow::on_actionExportSVG_triggered() {
  if (drawArea->getNet()) {
    QString fileName(lastExportPath);
    if (!fileName.endsWith('/')) {
      fileName += "/";
    }
    QFileInfo fic(currentFile);
    fileName += fic.baseName() + ".svg";
    fileName = QFileDialog::getSaveFileName(this,
                                            "Export SVG",
                                            fileName,
                                            "Scalable Vector Graphics (*.svg)");
    if (fileName.isEmpty()) {
      return;
    }
    
    bool exportOk = Export::saveSvg(drawArea->getNet(), &drawProps, fileName);
    
    // Remember last export path
    QFileInfo fi(fileName);
    lastExportPath = fi.absolutePath();
    
    // Show message for 2 sec.
    if (exportOk) {
      statusBar()->showMessage("Image exported", 2000);
    } else {
      QMessageBox::warning(this, "PNML view",
                           QString("Exporting to file ``%1'' failed.")
                           .arg(fileName));
    }
  }
}
#endif


void MainWindow::on_actionExportRaster_triggered() {
  if (drawArea->getNet()) {
    QString types;
    bool first = true;
    foreach (QByteArray format, QImageWriter::supportedImageFormats()) {
      if (format.toLower() == "mng" ||
          format.toLower() == "jpeg") {
        // MNG is for multiple pictures and JPG is already included in the
        // list, so JPEG is not needed
        continue;
      }
      types += QString(first?"":";;") +
                QString("%1 image (*.%2)")
                  .arg(QString(format).toUpper())
                  .arg(QString(format).toLower());
      first = false;
    }
    QString fileName(lastExportPath);
    if (!fileName.endsWith('/')) {
      fileName += "/";
    }
    QFileInfo fic(currentFile);
    fileName += fic.baseName() + "." + lastImageFormat.toLower();
    QString fileType = lastImageFormat.toUpper() + " image (*." + lastImageFormat.toLower() + ")";
    fileName = QFileDialog::getSaveFileName(this,
                                            "Export raster image",
                                            fileName,
                                            types,
                                            &fileType);
    if (fileName.isEmpty()) {
      return;
    }
    
    lastImageFormat = fileType.left(fileType.indexOf(' '));
    bool exportOk = Export::saveRaster(drawArea->getNet(), &drawProps, fileName, lastImageFormat);
    
    // Remember stuff for next time
    QFileInfo fi(fileName);
    lastExportPath = fi.absolutePath();
    
    // Show message for 2 sec.
    if (exportOk) {
      statusBar()->showMessage("Image exported", 2000);
    } else {
      QMessageBox::warning(this, "PNML view",
                           QString("Exporting to file ``%1'' failed.")
                           .arg(fileName));
    }
  }
}

void MainWindow::on_actionBatchConvert_triggered() {
  BatchConvertDialog dlg(&drawProps, this);
  dlg.exec();
}

void MainWindow::on_actionZoomIn_triggered() {
  if (drawProps.zoom() * 1.5 < 10) {
    drawProps.setZoom(drawProps.zoom() * 1.5);
    if (net) {
      net->setDirty(Net::DirtyZoom);
    }
    drawArea->sizeChanged();
  }
}

void MainWindow::on_actionZoomOut_triggered() {
  if (drawProps.zoom() / 1.5 > 0.1) {
    drawProps.setZoom(drawProps.zoom() / 1.5);
    if (net) {
      net->setDirty(Net::DirtyZoom);
    }
    drawArea->sizeChanged();
  }
}

void MainWindow::on_actionZoomFull_triggered() {
  drawProps.setZoom(1.0);
  if (net) {
    net->setDirty(Net::DirtyZoom);
  }
  drawArea->sizeChanged();
}

void MainWindow::on_actionAbout_triggered() {
  QString versionInfo("Version %1");
  versionInfo = versionInfo.arg(PNMLVIEW_VERSION);
#ifdef PNMLVIEW_VERSION_SVN
  QString svnVersion(PNMLVIEW_SVN_REV);
  svnVersion.replace("$", "");
  svnVersion.replace("Revision:", "");
  svnVersion = svnVersion.trimmed();
  versionInfo.append("_svn<br />\nSVN revision: %1");
  versionInfo = versionInfo.arg(svnVersion);
#endif
  
  QDialog dialog(this);
  Ui::VersionDialog vd;
  vd.setupUi(&dialog);
  vd.versionLabel->setText(versionInfo);
  dialog.exec();
}

void MainWindow::on_actionAboutQt_triggered() {
  QMessageBox::aboutQt(this);
}

void MainWindow::on_namesCheckBox_toggled(bool checked) {
  drawProps.setOption("ShowNames", checked);
  if (net) {
    net->setDirty(Net::DirtyLabels);
  }
  drawArea->update();
}

void MainWindow::on_rolesCheckBox_toggled(bool checked) {
  drawProps.setOption("ShowRoles", checked);
  if (net) {
    net->setDirty(Net::DirtyLabels);
  }
  drawArea->update();
}

void MainWindow::on_toolTipsCheckBox_toggled(bool checked) {
  drawArea->setToolTips(checked);
}

// Open some subnet if user clicks on it in hierarchy view
void MainWindow::on_hierarchyTreeWidget_itemClicked(QTreeWidgetItem *item, int column) {
  QVariant varNet = item->data(column, Qt::UserRole);
  if (varNet.canConvert<void *>()) {
    Net *newNet = static_cast<Net *>(varNet.value<void *>());
    if (newNet) {
      drawArea->setNet(newNet);
    }
  }
}
