/*
 * FILE:    net.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "net.h"
#include "transition.h"
#include "arc.h"
#include "place.h"
#include "page.h"
#include "comment.h"
#include "drawprops.h"

#include <QPainter>
#include <QTreeWidget>
#include <QPixmapCache>

#define MARGIN 20
#define CACHE_SIZE 5120 // 5MB pixmap cache

Net::Net () {
  parent = NULL;
  dirtyFlags = Net::AllDirty;
  QPixmapCache::setCacheLimit(CACHE_SIZE);
}

Net::~Net () {
//  qDebug("Net::~Net()");
  
  // Not necessary, but ok
  QPixmapCache::remove(key());
  
  Connectable *curCon;
  foreach (curCon, connectables) {
    delete curCon;
  }
  
  Arc *curArc;
  foreach (curArc, arcs) {
    delete curArc;
  }
  
  Comment *curComment;
  foreach (curComment, comments) {
    delete curComment;
  }
}

// Pre: pd, drawProps
void Net::draw (QPaintDevice *pd, DrawProps *drawProps, bool bypassCache /*= false*/) {
  qDebug("Net::draw()");
  
  if (bypassCache) {
    // bypassCache should also re-place the labels, according to current
    // font metrics for the given QPaintDevice. This is done after the
    // painter is created below (using solveDirty())
    setDirty(DirtyLabels);
  } else {
    solveDirty(drawProps);
  }
  
  if (pd->width() * pd->height() * 4 / 1024 > CACHE_SIZE) {
    // Don't draw big pixmaps
    bypassCache = true;
  }
  
  QRectF rect(0, 0, pd->width(), pd->height());
  QPixmap pixmap;
  if (bypassCache || !QPixmapCache::find(key(), pixmap)) {
    //qDebug("Net::draw() image not in cache or cache bypassed");
    
    QPainter painter;
    if (bypassCache) {
      // Warning! If using windows and printing to file,
      // the UI may get update events and calls this function again,
      // resulting in two draw()s running at the same time...
      if (!painter.begin(pd)) {
      	return;
      }
      solveDirty(drawProps, &painter);
      
      // Workaround for Windows clipping problem
      //   (http://www.trolltech.com/developer/tasktracker.html?method=entry&id=93953)
#ifdef Q_OS_WIN32
      if (QString(qVersion()) == "4.1.0" && !drawProps->updateRegion.isNull()) {
        painter.setClipRegion(drawProps->updateRegion);
      }
#endif
    } else {
      pixmap = QPixmap(pd->width(), pd->height());
      painter.begin(&pixmap);
    }
    
    painter.setRenderHint(QPainter::Antialiasing, drawProps->option("AntiAlias"));
    painter.setRenderHint(QPainter::TextAntialiasing, drawProps->option("AntiAlias"));
    
    painter.fillRect(rect, drawProps->brush("Background"));
    
    // Setup world
    QMatrix world;
    world.scale(drawProps->zoom(), drawProps->zoom());
    // 0.5 is added in the next line to draw sharper axes-parallel lines
    // if anti-aliasing is enabled
    world.translate(MARGIN - boundingBox.left() + 0.5, MARGIN - boundingBox.top() + 0.5);
    painter.setMatrix(world); // Enable world translation/scaling

    // Draw all connectables in this net
    Connectable *curCon;
    foreach (curCon, connectables) {
      curCon->draw(&painter, drawProps);
    }

    // Draw all arcs in this net
    // NB: if we would draw the arcs first and the connectables later
    //     and not using transparant shadow (such as when generating pdf files),
    //     the shadow would overlap the arcs.
    Arc *curArc;
    foreach (curArc, arcs) {
      curArc->draw(&painter, drawProps);
    }

    // Don't have to do draw the comments because they're labels

    // Draw the labels
    labelPlacer.draw(&painter, drawProps);
    
    if (!bypassCache) {
      QPixmapCache::insert(key(), pixmap);
    }
    painter.end();
  }
  
  
  if (!bypassCache) {
    QPainter painter(pd);
    
    // Workaround for Windows clipping problem
    //   (http://www.trolltech.com/developer/tasktracker.html?method=entry&id=93953)
#ifdef Q_OS_WIN32
    if (!drawProps->updateRegion.isNull()) {
      painter.setClipRegion(drawProps->updateRegion);
    }
#endif
    
    // Draw (cached) pixmap to screen
    painter.drawPixmap(rect, pixmap, rect);
  } else {
    // Because the labels are placed according to the current font metrics
    // of the given QPaintDevice, they should be recalculated next draw
    setDirty(DirtyLabels);
  }
}

QString Net::key() const {
  QString result;
  result.sprintf("%p", static_cast<const void *>(this));
  return result;
}

QSize Net::getSize(DrawProps *drawProps) {
  solveDirty(drawProps);
  return (boundingBox.size() + QSize(MARGIN*2, MARGIN*2)) * drawProps->zoom();
}

QVector<Page *> Net::getPages() const {
  return pages;
}

QVector<Connectable *> Net::getConnectables() const {
  return connectables;
}

QVector<Arc *> Net::getArcs() const {
  return arcs;
}

Connectable *Net::findConnectableById(QString id) const {
  Connectable *curCon;
  foreach (curCon, connectables) {
    if (curCon->getId() == id) {
      return curCon;
    }
  }
  return NULL;
}

// Pre: root
void Net::updateHierarchy(QTreeWidgetItem *root) const {
  // Add pages of current net as childnodes to root
  Page *curPage;
  foreach (curPage, pages) {
    if (curPage->getNet()) {
      QTreeWidgetItem *item = new QTreeWidgetItem(root);
      if (curPage->getName().isNull()) {
        item->setText(0, curPage->getId());
      } else {
        item->setText(0, curPage->getName());
      }
      //tree->setItemExpanded(item, true);
      QVariant varNet;
      varNet.setValue<void *>(static_cast<void *>(curPage->getNet()));
      item->setData(0, Qt::UserRole, varNet);
      if (curPage->getNet()) {
        curPage->getNet()->updateHierarchy(item);
      }
    }
  }
}

void Net::setDirty(Net::DirtyFlags df, bool recursive /*= true*/) {
  //qDebug("Net::setDirty()");
  
  dirtyFlags |= df;
  
  if (recursive) {
    // Update child nets
    Page *curPage;
    foreach (curPage, pages) {
      if (curPage->getNet()) {
        curPage->getNet()->setDirty(df, true);
      }
    }
  }
}

// Pre: drawProps && (painter ? painter.isActive() : true)
void Net::solveDirty(DrawProps *drawProps, QPainter *painter /*= NULL*/) {
  if (dirtyFlags & DirtyDrawables) {
    dirtyFlags |= DirtyLabels;
    labelPlacer.clearAll();
    Arc *arc;
    foreach (arc, arcs) {
      arc->updateLabels(&labelPlacer, drawProps);
    }
    Connectable *con;
    foreach (con, connectables) {
      con->updateLabels(&labelPlacer, drawProps);
    }
    Comment *comment;
    foreach (comment, comments) {
      comment->updateLabels(&labelPlacer, drawProps);
    }
  }
  if (dirtyFlags & DirtyLabels) {
    labelPlacer.setChanged();
  }
  if (dirtyFlags & DirtyDrawables ||
      dirtyFlags & DirtyLabels) {
    // Update bounding box
    boundingBox = QRect();
    Arc *arc;
    foreach (arc, arcs) {
      boundingBox |= arc->getBoundingRect(); // Unite rects
    }
    Connectable *con;
    foreach (con, connectables) {
      boundingBox |= con->getBoundingRect(); // Unite rects
    }
    boundingBox |= labelPlacer.getBoundingRect(drawProps, painter);
  }
  
  if (dirtyFlags) {
    QPixmapCache::remove(key());
  }
  dirtyFlags = 0;
}

QPoint Net::demapPos(QPoint pos, DrawProps *drawProps) {
  solveDirty(drawProps);
  
  // Inverse of code in Net::draw()
  return (pos / drawProps->zoom()) - QPoint(MARGIN - boundingBox.left(),
                                            MARGIN - boundingBox.top());
}

void Net::addConnectable(Connectable * con) {
  if (!con) { return; }
  
  connectables << con;
  setDirty(Net::DirtyDrawables, false);
}

void Net::addArc(Arc *newArc) {
  if (!newArc) { return; }
 
  bool found = false;
  if (newArc->getSource() && newArc->getTarget()) {
    const QVector<Arc *> arcs = newArc->getSource()->getArcs();
    foreach (Arc *arc, arcs) {
      if (newArc->getTarget() == arc->getTarget()) {
        // { newArc->getSource() == arc->getSource() }
        if (arc->getPoints() == newArc->getPoints() &&
            arc->getType() == newArc->getType()) {
          arc->merge(newArc, false);
          delete newArc;
          newArc = NULL;
          found = true;
          break;
        }
      } else if (newArc->getTarget() == arc->getSource()) {
        // { newArc->getSource() == arc->getTarget() }
        if (arc->getPoints() == newArc->getPoints() &&
            arc->getType() == newArc->getType()) {
          arc->merge(newArc, true);
          delete newArc;
          newArc = NULL;
          found = true;
          break;
        }
      }
    }
  }
  
  if (!found) {
    if (newArc->getSource()) {
      newArc->getSource()->addArc(newArc);
    }
    if (newArc->getTarget()) {
      newArc->getTarget()->addArc(newArc);
    }
    arcs << newArc;
  }
  
  setDirty(Net::DirtyDrawables, false);
}

void Net::addComment(Comment * comment) {
  if (!comment) { return; }
  
  comments << comment;
  setDirty(Net::DirtyDrawables, false);
}

void Net::addPage(Page * page) {
  if (!page) { return; }
  
  // Note: pages are also kept in this separate vector to be able to
  //       quickly find subnets
  pages << page;
}

bool Net::isEmpty() {
  return connectables.isEmpty() &&
         arcs.isEmpty() &&
         comments.isEmpty();
}
