/*
 * FILE:    pnmlhandler.cpp
 * PACKAGE: PNML view
 * LICENCE: See LICENCE file
 * AUTHOR:  Freek van Walderveen
 *
 */

#include "pnmlhandler.h"
#include "transition.h"
#include "place.h"
#include "arc.h"
#include "page.h"
#include "comment.h"

#include <QMessageBox>

#define CURRENTNET pageStack.top()->getNet()

PnmlHandler::PnmlHandler() {
  metPnmlTag = false;
  readingText = false;
  net = NULL;
  curObject = none;
  curPlace = NULL;
  curArc = NULL;
  curTrans = NULL;
}

PnmlHandler::~PnmlHandler() {
  if (!pageStack.isEmpty()) {
    pageStack.top()->setNet(NULL);
    delete pageStack.top();
  }
}

bool PnmlHandler::startElement(const QString & /* namespaceURI */,
                               const QString & /* localName */,
                               const QString &qName,
                               const QXmlAttributes &attributes) {
  QString lcName = qName.toLower();
  if (!metPnmlTag && lcName != "pnml") {
    errorStr = "The file is not a PNML file.";
    return false;
  }
  
  // Remember tag names, but forget about "toolspecific", just read
  //   its subtags
  if (lcName != "toolspecific") {
    tagStack.push(lcName);
  }

  if (lcName == "pnml") {
    metPnmlTag = true;
  } else if (lcName == "net") {
    net = new Net;
    // Make a dummy page on the stack
    Page *dummyPage = new Page;
    dummyPage->setNet(net);
    pageStack.push(dummyPage);
  } else if (lcName == "position") {
    if (tagStack.size() < 2) { return true; }
    
    QPoint point;
    point.setX(attributes.value("x").toInt());
    point.setY(attributes.value("y").toInt());
    if (tagStack.at(tagStack.size()-2) == "graphics") {
      switch (curObject) {
        case transition:
          if (curTrans) {
            curTrans->setPosition(point);
          } else {
            return false;
          }
          break;
        case place:
          if (curPlace) {
            curPlace->setPosition(point);
          } else {
            return false;
          }
          break;
        case page:
          if (!pageStack.isEmpty()) {
            pageStack.top()->setPosition(point);
          } else {
            return false;
          }
          break;
        case arc:
          if (curArc) {
            curArc->addPoint(point);
          } else {
            return false;
          }
          break;
        case comment:
          if (curComment) {
            curComment->setPosition(point);
          } else {
            return false;
          }
          break;
        default:
          break;
      }
    }
#ifdef SHOW_FAKE_ARCS
      else if (tagStack.at(tagStack.size()-2) == "pathgraphics") {
      switch (curObject) {
        case place:
          if (curPlace) {
            curPlace->pathGraphics->append(point);
          } else {
            return false;
          }
          break;
        default:
          break;
      }
    }
#endif
  } else if (lcName == "dimension") {
    int x = attributes.value("x").toInt();
    int y = attributes.value("y").toInt();
    if (x >= 0 && y >= 0) {
      QSize size(x, y);
      switch (curObject) {
        case transition:
          if (curTrans) {
            curTrans->setSize(size);
          } else {
            return false;
          }
          break;
        case place:
          if (curPlace) {
            curPlace->setSize(size);
          } else {
            return false;
          }
          break;
        case page:
          if (!pageStack.isEmpty()) {
            pageStack.top()->setSize(size);
          } else {
            return false;
          }
          break;
        case comment:
          if (curComment) {
            curComment->setSize(size);
          } else {
            return false;
          }
          break;
        default:
          qWarning("dimension not supported for this objecttype");
          break;
      }
    }
  } else if (lcName == "transition") {
    curTrans = new Transition;
    curTrans->setId(attributes.value("id"));
    curObject = transition;
  } else if (lcName == "place") {
    curPlace = new Place;
    curPlace->setId(attributes.value("id"));
    curObject = place;
  } else if (lcName == "referenceplace") {
    curPlace = new Place;
    curPlace->setType(Place::Reference);
    
    // Find the refering place in the parent net
    if (!pageStack.isEmpty() && CURRENTNET) {
      Net *parentNet = CURRENTNET->getParent();
      Connectable *reference = NULL;
      if (parentNet) {
        reference = parentNet->findConnectableById(attributes.value("ref"));
        if (reference && reference->getConType() == Connectable::Place) {
          curPlace->setReference(dynamic_cast<Place *>(reference));
        } else {
          reference = NULL;
        }
      }
      if (!reference) {
        errorStr = "referencePlace without parent";
        return false;
      }
    } else {
        errorStr = "No current net?";
        return false;
    }
    
    curPlace->setId(attributes.value("id"));
    curObject = place;
  } else if (lcName == "page") {
    if (pageStack.isEmpty() || !CURRENTNET) {
      errorStr = "page without net";
      return false;
    }
    
    Page *curPage = new Page;
    Net *newNet = new Net;
    newNet->setParent(CURRENTNET);
    curPage->setNet(newNet);
    curPage->setId(attributes.value("id"));
    pageStack.push(curPage);
    curObject = page;
  } else if (lcName == "arc") {
    if (pageStack.isEmpty() || !CURRENTNET) {
      errorStr = "arc without net";
      return false;
    }
    
    curArc = new Arc;
    curArc->setId(attributes.value("id"));
    
    Connectable *source, *target;
    source = CURRENTNET->findConnectableById(attributes.value("source"));
    target = CURRENTNET->findConnectableById(attributes.value("target"));
    curArc->setSource(source);
    curArc->setTarget(target);
    curArc->updateConsumption(1);
    curObject = arc;
  } else if (lcName == "comment") {
    curComment = new Comment;
    curComment->setId(attributes.value("id"));
    curObject = comment;
  } else if (lcName == "role") {
    curRoleId = attributes.value("id");
    if (curRoleId.isEmpty()) {
      curRoleId = attributes.value("ref");
    }
    if (curObject == none) {
      curObject = role;
    }
#ifdef SHOW_FAKE_ARCS
  } else if (lcName == "pathgraphics") {
    switch (curObject) {
      case place:
        if (curPlace) {
          curPlace->pathGraphics = new QVector<QPoint>;
        } else {
          return false;
        }
        break;
      default:
        qWarning("pathgraphics not supported for this objecttype");
        break;
    }
#endif
  } else if (lcName == "text") {
    readingText = true;
    currentText.clear();
  } else if (lcName == "graphics" ||
             lcName == "name" ||
             lcName == "toolspecific" ||
             lcName == "tokencasesensitive" ||
             lcName == "type" ||
             lcName == "prequery" ||
             lcName == "postquery" ||
             lcName == "description" ||
             lcName == "collector" ||
             lcName == "emitter" ||
             lcName == "emitor" ||
             lcName == "actor" ||
             lcName == "cost" ||
             lcName == "processingtime" ||
             lcName == "fixed" ||
             lcName == "variable" ||
             lcName == "mean" ||
             lcName == "deviation" ||
             lcName == "inline" ||
             lcName == "initialmarking" ||
             lcName == "roles") {
    // ignoring these tags as start-element
  } else {
    qWarning("Unhandled start-element: %s", qPrintable(qName));
  }

  return true;
}

bool PnmlHandler::endElement(const QString & /* namespaceURI */,
                             const QString & /* localName */,
                             const QString &qName) {
  QString lcName = qName.toLower();
  if (lcName != "toolspecific") {
    if (tagStack.isEmpty()) {
      errorStr = "close tag without start tag?";
      return false;
    }
    tagStack.pop();
  }
  
  if (lcName == "transition") {
    if (!pageStack.isEmpty() && CURRENTNET) {
      CURRENTNET->addConnectable(curTrans);
      curTrans = NULL;
    } else {
      return false;
    }
    curObject = none;
  } else if (lcName == "place") {
    if (!pageStack.isEmpty() && CURRENTNET) {
      CURRENTNET->addConnectable(curPlace);
      curPlace = NULL;
    } else {
      return false;
    }
    curObject = none;
  } else if (lcName == "referenceplace") {
    if (!pageStack.isEmpty() && CURRENTNET) {
      CURRENTNET->addConnectable(curPlace);
      curPlace = NULL;
    } else {
      return false;
    }
    curObject = none;
  } else if (lcName == "page") {
    if (pageStack.isEmpty() || !CURRENTNET) {
      errorStr = "page without parent";
      return false;
    }
    
    Page *childPage = pageStack.pop();
    CURRENTNET->addPage(childPage);
    CURRENTNET->addConnectable(childPage);
    curObject = none;
  } else if (lcName == "comment") {
    if (!pageStack.isEmpty() && CURRENTNET) {
      CURRENTNET->addComment(curComment);
      curComment = NULL;
    } else {
      return false;
    }
    curObject = none;
  } else if (lcName == "role") {
    if (!curRoleId.isEmpty()) {
      switch (curObject) {
        case transition:
          if (!pageStack.isEmpty() && pageStack.first()->getNet() && curTrans) {
            curTrans->addRole(pageStack.first()->getNet()->roles[curRoleId]);
          }
          break;
        default:
          break;
      }
      
      curRoleId = "";
    }
  } else if (lcName == "collector") {
    switch (curObject) {
      case transition:
        if (!pageStack.isEmpty() && CURRENTNET && curTrans) {
          if (currentText.trimmed().toLower() == "true") {
            curTrans->setCollector(true);
          } else {
            curTrans->setCollector(false);
          }
        } else {
          return false;
        }
        break;
      default:
        break;
    }
  } else if (lcName == "emitter" || // Just in case Yasper changes its output
             lcName == "emitor") {
    switch (curObject) {
      case transition:
        if (!pageStack.isEmpty() && CURRENTNET && curTrans) {
          if (currentText.trimmed().toLower() == "true") {
            curTrans->setEmitter(true);
          } else {
            curTrans->setEmitter(false);
          }
        } else {
          return false;
        }
        break;
      default:
        break;
    }
  } else if (lcName == "inline") {
    switch (curObject) {
      case page:
        if (!pageStack.isEmpty()) {
          if (currentText.trimmed().toLower() == "true") {
            pageStack.top()->setType(Transition::InlinePage);
          } else {
            pageStack.top()->setType(Transition::Page);
          }
        } else {
          return false;
        }
        break;
      default:
        break;
    }
  } else if (lcName == "arc") {
    if (!pageStack.isEmpty() && CURRENTNET && curArc) {
#ifdef SHOW_FAKE_ARCS
      // Check for involvement of reference places
      Arc *fakeArc = NULL;
      Connectable *source = curArc->getSource();
      Connectable *target = curArc->getTarget();
      if (source && source->getConType() == Connectable::Place) {
        Place *sourceRef = dynamic_cast<Place *>(source);
        int depth = 0;
        while (sourceRef && sourceRef->getType() == Place::Reference && pageStack.size() > 1 + depth) {
          // Create a "fake" arc in the parent net
          fakeArc = new Arc;
          fakeArc->setId(curArc->getId() + "_fake" + QString::number(depth));
          fakeArc->setType(curArc->getType());
          fakeArc->updateConsumption(curArc->getConsumption());
          fakeArc->updateProduction(curArc->getProduction());
          fakeArc->setStoreAction(curArc->getStoreAction());
          if (sourceRef->pathGraphics) {
            fakeArc->setPoints(*(sourceRef->pathGraphics));
          }
          if (!sourceRef->getReference()) {
            qDebug("No reference for referencePlace ``%s''?", qPrintable(sourceRef->getId()));
          }
          fakeArc->setSource(sourceRef->getReference());
          fakeArc->setTarget(pageStack.at(pageStack.size() - 1 - depth));

          // Set reference type of sourceRef
          if (curArc->getType() == Arc::Standard && 
              (sourceRef->getRefType() == Place::OutFlowRef ||
               sourceRef->getRefType() == Place::BiFlowRef)) {
            sourceRef->setRefType(Place::BiFlowRef);
          } else {
            switch (curArc->getType()) {
              case Arc::Standard:
                sourceRef->setRefType(Place::InFlowRef);
                break;
              case Arc::Reset:
                sourceRef->setRefType(Place::ResetRef);
                break;
              case Arc::Inhibitor:
                sourceRef->setRefType(Place::InhibitorRef);
                break;
              case Arc::StoreAction:
                sourceRef->setRefType(Place::StoreRef);
                break;
            }
          }

          if (pageStack.at(pageStack.size() - 2 - depth)->getNet()) {
            pageStack.at(pageStack.size() - 2 - depth)->getNet()->addArc(fakeArc);
          }

          // For next iteration
          sourceRef = sourceRef->getReference();
          depth++;
        }
      } else if (target && target->getConType() == Connectable::Place) {
        Place *targetRef = dynamic_cast<Place *>(target);
        int depth = 0;
        while (targetRef && targetRef->getType() == Place::Reference && pageStack.size() > 1 + depth) {
          // Create a "fake" arc in the parent net
          fakeArc = new Arc;
          fakeArc->setId(curArc->getId() + "_fake" + QString::number(depth));
          fakeArc->setType(curArc->getType());
          fakeArc->updateConsumption(curArc->getConsumption());
          fakeArc->updateProduction(curArc->getProduction());
          fakeArc->setStoreAction(curArc->getStoreAction());
          if (targetRef->pathGraphics) {
            fakeArc->setPoints(*(targetRef->pathGraphics));
          }
          if (!targetRef->getReference()) {
            qDebug("No reference for referencePlace ``%s''?", qPrintable(targetRef->getId()));
          }
          fakeArc->setSource(pageStack.at(pageStack.size() - 1 - depth));
          fakeArc->setTarget(targetRef->getReference());

          // Set reference type of sourceRef
          if (curArc->getType() == Arc::Standard && 
              (targetRef->getRefType() == Place::InFlowRef ||
               targetRef->getRefType() == Place::BiFlowRef)) {
            targetRef->setRefType(Place::BiFlowRef);
          } else {
            switch (curArc->getType()) {
              case Arc::Standard:
                targetRef->setRefType(Place::OutFlowRef);
                break;
              case Arc::Reset:
                targetRef->setRefType(Place::ResetRef);
                break;
              case Arc::Inhibitor:
                targetRef->setRefType(Place::InhibitorRef);
                break;
              case Arc::StoreAction:
                targetRef->setRefType(Place::StoreRef);
                break;
            }
          }

          if (pageStack.at(pageStack.size() - 2 - depth)->getNet()) {
            pageStack.at(pageStack.size() - 2 - depth)->getNet()->addArc(fakeArc);
          }

          // For next iteration
          targetRef = targetRef->getReference();
          depth++;
        }
      }
#endif
      
      CURRENTNET->addArc(curArc);
      curArc = NULL;
    } else {
      return false;
    }
    curObject = none;
  } else if (lcName == "type") {
    switch (curObject) {
      case arc:
        if (curArc) {
          currentText = currentText.trimmed().toLower();
          if (currentText == "reset") {
            curArc->setType(Arc::Reset);
          } else if (currentText == "inhibitor") {
            curArc->setType(Arc::Inhibitor);
          } else {
            if (currentText.contains("c")) {
              curArc->setType(Arc::StoreAction);
              curArc->setStoreAction(Arc::ActionC);
            }
            if (currentText.contains("r")) {
              curArc->setType(Arc::StoreAction);
              curArc->setStoreAction(Arc::ActionR);
            }
            if (currentText.contains("u")) {
              curArc->setType(Arc::StoreAction);
              curArc->setStoreAction(Arc::ActionU);
            }
            if (currentText.contains("d")) {
              curArc->setType(Arc::StoreAction);
              curArc->setStoreAction(Arc::ActionD);
            }
          }
        }
        break;
      case transition:
        if (curTrans) {
          currentText = currentText.trimmed().toLower();
          if (currentText == "xor") {
            curTrans->setType(Transition::Xor);
          } else if (currentText == "and") {
            curTrans->setType(Transition::And);
          } else {
            qWarning("Type ``%s'' not implemented for transitions", qPrintable(currentText));
          }
        }
        break;
      case place:
        if (curPlace) {
          currentText = currentText.trimmed().toLower();
          if (currentText == "store") {
            curPlace->setType(Place::Store);
          } else if (currentText == "channel") {
            curPlace->setType(Place::Channel);
          } else {
            qWarning("Type ``%s'' not implemented for places", qPrintable(currentText));
          }
        }
        break;
      default:
        break;
    }
  } else if (lcName == "name") {
    if (tagStack.isEmpty()) {
      errorStr = "name without object";
      return false;
    }
    
    currentText = currentText.trimmed();
    switch (curObject) {
      case transition:
        if (curTrans && tagStack.top() == "transition") {
          curTrans->setName(currentText);
        }
        break;
      case place:
        if (curPlace &&
            (tagStack.top() == "place" ||
             tagStack.top() == "referenceplace")) {
          curPlace->setName(currentText);
        }
        break;
      case arc:
        if (curArc && tagStack.top() == "arc") {
          curArc->setName(currentText);
        }
        break;
      case page:
        if (!pageStack.isEmpty() && tagStack.top() == "page") {
          pageStack.top()->setName(currentText);
        }
        break;
      case role:
        if (!pageStack.isEmpty() && tagStack.top() == "role") {
          if (!currentText.isEmpty()) {
            // TODO: Are roles always global?
            if (pageStack.first()->getNet()) {
              pageStack.first()->getNet()->roles[curRoleId] = currentText;
            } else {
              errorStr = "rolename without net";
              return false;
            }
          }
        }
        break;
      case comment:
        //TODO Do comments have names?
        break;
      case none:
        // Set global netname
        if (!pageStack.isEmpty() && tagStack.top() == "net") {
          pageStack.first()->setName(currentText);
        }
        break;
    }
  } else if (lcName == "tokencasesensitive") {
    switch (curObject) {
      case place:
        if (tagStack.isEmpty()) { return false; }
        if (curPlace &&
            (tagStack.top() == "place" ||
             tagStack.top() == "referenceplace")) {
          if (currentText.trimmed().toLower() == "true") {
            curPlace->setCaseSensitive(Place::TokenCaseSensitive);
          } else {
            curPlace->setCaseSensitive(Place::TokenCaseInsensitive);
          }
        }
        break;
      default:
        qWarning("tokenCaseSensitive found outside place or store?");
        break;
    }
  } else if (lcName == "description") {
    if (tagStack.isEmpty()) { return false; }
    currentText = currentText.trimmed();
    switch (curObject) {
      case comment:
        if (curComment && tagStack.top() == "comment") {
          curComment->setDescription(currentText);
        }
        break;
      case place:
        if (curPlace &&
            (tagStack.top() == "place" ||
             tagStack.top() == "referenceplace")) {
          curPlace->setDescription(currentText);
        }
        break;
      case transition:
        if (curTrans && tagStack.top() == "transition") {
          curTrans->setDescription(currentText);
        }
        break;
      case page:
        if (!pageStack.isEmpty() && tagStack.top() == "page") {
          pageStack.top()->setDescription(currentText);
        }
        break;
      case arc:
      case role:
        qWarning("Description tag not implemented for this object type");
        break;
      case none:
        qWarning("Description tag without context?");
        break;
    }
  } else if (lcName == "prequery") {
    if (tagStack.isEmpty()) { return false; }
    switch (curObject) {
      case transition:
        if (curTrans && tagStack.top() == "transition") {
          curTrans->setPreQuery(currentText.trimmed());
        }
        break;
      default:
        qWarning("prequery found in non-transition object?");
        break;
    }
  } else if (lcName == "postquery") {
    if (tagStack.isEmpty()) { return false; }
    switch (curObject) {
      case transition:
        if (curTrans && tagStack.top() == "transition") {
          curTrans->setPostQuery(currentText.trimmed());
        }
        break;
      default:
        qWarning("postquery found in non-transition object?");
        break;
    }
  } else if (lcName == "mean" ||
             lcName == "deviation" ||
             lcName == "fixed" ||
             lcName == "variable") {
    if (tagStack.isEmpty()) { return false; }
    switch (curObject) {
      case transition:
        if (curTrans) {
          int value = currentText.trimmed().toInt();
          if (value > 0) {
            if (lcName == "mean" && tagStack.top() == "processingtime") {
              curTrans->setTimeMean(value);
            } else if (lcName == "deviation" && tagStack.top() == "processingtime") {
              curTrans->setTimeDeviation(value);
            } else if (lcName == "fixed" && tagStack.top() == "cost") {
              curTrans->setFixedCost(value);
            } else if (lcName == "variable" && tagStack.top() == "cost") {
              curTrans->setVariableCost(value);
            }
          }
        }
        break;
      default:
        qWarning("postquery found in non-transition object?");
        break;
    }
  } else if (lcName == "initialmarking") {
    if (tagStack.isEmpty()) { return false; }
    switch (curObject) {
      case place:
        if (curPlace && tagStack.top() == "place") {
          int marking = currentText.trimmed().toInt();
          if (marking < 0) { marking = 0; }
          curPlace->setInitialMarking(marking);
        }
        break;
      default:
        qWarning("initialMarking found in non-place object?");
        break;
    }
  } else if (lcName == "text") {
    readingText = false;
  } else if (lcName == "graphics" ||
             lcName == "pathgraphics" ||
             lcName == "position" ||
             lcName == "dimension" ||
             lcName == "net" ||
             lcName == "pnml" ||
             lcName == "actor" ||
             lcName == "roles" ||
             lcName == "cost" ||
             lcName == "processingtime" ||
             lcName == "toolspecific") {
    // ignoring these tags as end-element
  } else {
    qWarning("Unhandled end-element: %s", qPrintable(qName));
  }
  return true;
}

bool PnmlHandler::characters(const QString &str) {
  if (readingText) {
    currentText += str;
  }
  return true;
}

bool PnmlHandler::fatalError(const QXmlParseException &exception) {
  qWarning("Fatal PNML parse error");
  fatalErrorStr = QString("Parse error at line %1, column %2:\n%3")
                           .arg(exception.lineNumber())
                           .arg(exception.columnNumber())
                           .arg(exception.message());
  
  // TODO: clean up the mess (delete net, cur*)
  
  return false;
}

QString PnmlHandler::errorString() const {
  return errorStr;
}

QString PnmlHandler::fatalErrorString() const {
  return fatalErrorStr;
}

